from typing import Any

CODE_SIZE: int
MAGIC: int
copyright: str

def compile(*args, **kwargs) -> Any: ...
def fold_case(*args, **kwargs) -> Any: ...
def get_all_cases(*args, **kwargs) -> Any: ...
def get_code_size(*args, **kwargs) -> Any: ...
def get_expand_on_folding(*args, **kwargs) -> Any: ...
def get_properties(*args, **kwargs) -> Any: ...
def has_property_value(*args, **kwargs) -> Any: ...
